/* This is derived from NEWTRACK eyetracking software, (c) 2004 by */
/* Eric Auer. NEWTRACK is free software; you can redistribute it   */
/* and modify it under the terms of the GNU General Public License */
/* as published by the Free Software Foundation; either version 2  */
/* of the License, or (at your option) any later version.          */
/*     NEWTRACK is distributed in the hope that it will be useful, */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty */
/*     of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     */
/*     See the GNU General Public License for more details.        */
/* You should have received a copy of the GNU General Public       */
/* License (license.txt) along with this program; if not, check    */
/* www.gnu.org or write to the Free Software Foundation, Inc., 59  */
/* Temple Place, Suite 330, Boston, MA  02111-1307 USA.            */

/* lowlevel code for parallel port data acquisition module */

/* We connect afeed (inverted on sub-d) to all ~CS controls for  */
/* data from the module to the computer! Avoids port fighting... */

#include "main.h"    /* includes all the headers */

static int lpt_base = 0;	/* printer port if detected, -1 if disabled */
int tracking = 0;		/* 1 while tracking, set only by get_raw_xy */
int lpt_inverse = 0;		/* XOR mask for buttons at printer port */



/* find printer port device */
void init_lpt(void)
{
  int i;
  int x, y; /* x is not used here but will normally be 0..4095 or -1 */

  i = 0;
  do {
      lpt_base = peek (0x40, 8+i+i); /* check LPT(i-1) port */
      if (!lpt_base) {
	  i++;
	  y = 0;
      } else {
	  (void) get_raw_xy (&x, &y);
	  i++; /* do not get stuck! */
	  if ( (y<0) || (y>4095) ) {
	      y = 0; /* if bit stuck to 1, assume error */
	      out8 (lpt_base + 2, 0x0c);	/* normal mode again */
	  }
      }
  } while ( (i<3) && ( (!lpt_base) || (!y) ) );

  if (!y) {
      printf ("No printer port ADC device with non-zero Y coordinate found.\n");
      printf ("Using mouse / keyboard driver as fallback.\n");
      lpt_base = -1;
  } else {
      int key;

      printf ("    Printer port ADC device found at port %x. Press ESC to ignore it\n",
          lpt_base);
      printf ("    and use the mouse. Otherwise press ENTER to continue.\n");
      
      do {
	    key = getxkey ();
      } while ( (key != 27) && (key != 13) );   /* wait for response */
      
      if (key == 27) {				/* user wants only mice */
          printf ("  You selected using the mouse driver - ignoring ADC.\n");
          lpt_base = -1;			/* force mouse usage */
      }
  } /* device found */

} /* init_lpt */



uint32 get_raw_mouse(int * x, int * y) /* return raw mouse x/y coordinates */
{
  static int mickey_x = -1; /* mickey X coordinate mouse */
  static int mickey_y = -1; /* mickey Y coordinate mouse */
  static int mouse_buttons = 0;
  int86regs r;

#ifdef DEBUG_MOUSE_TIMING
  uint32 T;
#endif

  if (mickey_x < 0) {
      if ( (!peek (0, 0x33*4)) || (!peek (0, (0x33*4) + 2)) ) {
	  printf("You need a mouse driver now, sorry.\n");
	  exit (1);
      }
      r.w.ax = 0;       /* Function AX 0 is "init / start mouse driver" */
      (void) intr (0x33, &r, &r);
      if (r.w.ax != 0xffff) {
	  printf("Mouse driver not cooperating, sorry.\n");
	  exit (1);
      } else {
	  printf("Mouse with %d buttons found and activated.\n",
	    (r.w.bx == 0xffff) ? 2 : (int)r.w.bx );
      }
      mickey_x = 2047;
      mickey_y = 2047;
  } /* mouse initialization */

#ifdef DEBUG_MOUSE_TIMING
  T = get_ytime ();
#endif
  r.w.ax = 0x0b;        /* Function AX B is "get relative movement" */
  (void) intr (0x33, &r, &r);
  /* calling function B not only reads the value but also resets it */

  { /* update position, using mickeys */
      int delta_x = (sint16)(r.w.cx);
      int delta_y = (sint16)(r.w.dx);
      mickey_x += delta_x; /* movement unit is mickeys - fine grained */
      mickey_y += delta_y; /* movement unit is mickeys - fine grained */
      if ((mickey_x < 1) || (mickey_x > 4094))
	  mickey_x = (mickey_x < 1) ? 1 : 4094;
      if ((mickey_y < 1) || (mickey_y > 4094))
	  mickey_y = (mickey_y < 1) ? 1 : 4094;
  } /* position update */

#ifdef DEBUG_MOUSE_TIMING
  printf("int 33.b took %u ysec\n", get_ytime () - T);
#endif

  if (r.w.cx | r.w.dx) {

      /* This can be very SLOW because the driver waits for a data packet */
      /* So we only do it if the mouse has moved according to int 33.000b */
#ifdef DEBUG_MOUSE_TIMING
      T = get_ytime ();
#endif
      r.w.ax = 3;               /* Function AX 3 is "get position / buttons */
      (void) intr (0x33, &r, &r);
      mouse_buttons = r.w.bx & 3; /* only check left / right button */
#ifdef DEBUG_MOUSE_TIMING
      printf("int 33.3 took %u ysec\n", get_ytime () - T);
#endif
      tracking = (mouse_buttons & 3) ? 0 : 1;	/* click to get lost :-) */
  } /* check for mouse buttons */

  *x = mickey_x;
  *y = mickey_y;
  return get_ytime ();

} /* get_raw_mouse */



/* generate timing for LTC1286: CLK must stay stable at least 2ysec  */
/* and total clock frequency must not exceed 200 kHz. Given, because */
/* (2 * wait_2ysec) + (2 * write clock) + (1 * read data) with I/O   */
/* speeds of 1 ysec per LPT port I/O certainly takes that long.      */
void wait_2ysec(void)
{
  /* two possible versions: trust LPT port access to take > 1ysec */
  /* or use get_ytime() to be far on the safe side (wait 2.?ysec) */

#ifdef SAFE_WAIT_2YSEC

  uint32 first_time;
  first_time = get_ytime ();
  do {
    __asm__ __volatile__ ("inb $0x71, %%al\n" : : "a");
	/* we are bored and read the CMOS data port all the time */
  } while ((get_ytime () - first_time) < 2)

#else
   __asm__ __volatile__ (       /* volatile -> do not optimize */
       "inb %%dx, %%al\n"
       "inb %%dx, %%al\n"
       :                        /* output: nothing */
       : "d" (lpt_base)         /* input: port     */
       : "ax"                   /* modified: EAX   */
       );

#endif

} /* wait_2ysec */



uint32 get_raw_xy(int * x, int * y)  /* return raw x/y coordinates */
{   /* values will be in 0..4095 range or -1 for 12bit A/D module. */
    /* If track loss or blink, only x will be forced to -1...      */
  int bits, lost, lpt_control;
  uint32 sample_moment, time_check;
  int xtmp, ytmp, i;

  if (!lpt_base)                        /* A/D device not yet known? */
      init_lpt ();

  if (lpt_base < 0)                     /* A/D device simulation mode? */
      return get_raw_mouse (x, y);	/* use only mouse for simluation */

  lpt_control = in8 (lpt_base+2) | 0x20; /* always use direction: IN */
  lpt_control |= 1; /* set STB to true, to pull ~STB on port low!    */
      /* (the A/D chips want this on LOW during init: used as CLK!)  */

  time_check = get_ytime ();

  out8 (lpt_base+2, lpt_control & ~2);  /* DISABLE the chips */
      /* pull ~CS high for a while to initiate a new conversion cycle   */
      /* ~CS is connected to ~AFEED which is INVERSE of lpt_control & 2 */
      /* Only an disable -> enable transition starts a new conversion!  */
  wait_2ysec ();
  out8 (lpt_base+2, lpt_control | 2);   /* ENABLE all chips */
      /* let the fun begin! (important: printer port is in IN direction) */
  wait_2ysec ();
  bits = in8 (lpt_base+1);      /* get status port */
  lost = !(bits & TLBBIT);      /* check if tracking lost (or blink) */

  if ((get_ytime () - time_check) < 5) {
      printf ("Unexpected problem in acquire.c! LPT I/O takes < 1ysec.\n");
      printf ("Please recompile with alternative wait_2ysec version.\n");
      exit (1);
  }

  for (i = 0; i<2; i++) {
      (void) in8 (lpt_base);                    /* takes 1ysec itself */
	  /* for ISA or LPT ports, this takes 1.2ysec. Less for PCI.  */
      out8 (lpt_base+2, lpt_control & ~1);      /* let CLK rise */
      wait_2ysec ();
      out8 (lpt_base+2, lpt_control | 1);       /* let CLK fall */
      wait_2ysec ();
  } /* 2 clock cycles for sample and hold */
  
  sample_moment = get_ytime ();
  xtmp = 0;
  ytmp = 0;

  for (i = 4096; i; i>>=1) {
      bits = in8 (lpt_base);                    /* fetch actual data bits */
      if (bits & X_ADCBIT) xtmp |= i;
      if (bits & Y_ADCBIT) ytmp |= i;
      out8 (lpt_base+2, lpt_control & ~1);      /* let CLK rise */
      wait_2ysec ();
      out8 (lpt_base+2, lpt_control | 1);       /* let CLK fall */
      wait_2ysec ();
  } /* collect 12+1 bits: 0, then MSB ... LSB ... MSB, then zeroes */
  /* because of the MSB-LSB-MSB scheme, you can pick an endianness */

  for (i = 1; i<12; i++) {
      (void) in8 (lpt_base); /* could pick up bits 1 .. 11 again here */
      out8 (lpt_base+2, lpt_control & ~1);      /* let CLK rise */
      wait_2ysec ();
      out8 (lpt_base+2, lpt_control | 1);       /* let CLK fall */
      wait_2ysec ();
  } /* extra cycles just return 0 bits */
  /* we could pull ~CS high earlier and stop clocking earlier,  */
  /* but LTC1286 wants at least 80 ysec total cycle time anyway */

  /* keep CLK down - no need to pull up ~CS now. */
  *x = xtmp;
  *y = ytmp;

  if ((ytmp | xtmp) & 4096) {   /* zero bit missing - chip not there? */
      *x = -1;			/* signal reading error */
      *y = -1;
  } /* no chip found!? */

  tracking = !lost;		/* obvious... */

  out8 (lpt_base+2, 0x0c);	/* disable ADC device again: */
  				/* set port to normal output mode again. */

  return sample_moment;

} /* get_raw_xy */



/* If keyboard key, buttons will have key ASCII << 8 and full */
/* key << 16 ORed in. */
uint32 get_buttons(int * bits) /* return bitmask: pressed buttons */
{  /* low bits are combination of ...BUTTON values defined above. */
  int mybuttons, shift1, shift2;

  mybuttons = 0;

  if (!lpt_base)
      init_lpt ();

  if (kbhit ()) {       /* quick check whether key is waiting */
      int key;

      key = getxkey ();         /* adds 0x100 or 0x200 if specials */
	  /* getch() would use DOS */
      mybuttons |= key << 16;                           /* keyboard, verbose */
      switch (key) {
	  case  27: mybuttons |= ESCBUTTON; break;      /* ESC */
	  case   4:                                     /* ^D */
	  case   3: mybuttons |= ABORTBUTTON; break;    /* ^C */
	  case 'Y':
	  case 'y': mybuttons |= YESBUTTON; break;
	  case 'N':
	  case 'n': mybuttons |= NOBUTTON; break;
	  case 'C':
	  case 'c': mybuttons |= CALBUTTON; break;
	  case  13: mybuttons |= NEXTBUTTON; break;     /* ENTER */
	  default:  if (key < 0x100)
	      mybuttons |= key << 8;                    /* keyboard */
      } /* switch / case */
  } /* key waiting */


  if ( (lpt_base < 0) || (peekb (0x40,0x17) & 0x10) ) {
      /* no printer device found - or Scroll Lock active! */

      shift1 = peekb (0x40, 0x17);      /* 1st shift bit mask */
      shift2 = peekb (0x40, 0x18);      /* 2nd shift bit mask */

      if (shift1 & 2) mybuttons |= YESBUTTON; /* left  shift */
      if (shift1 & 1) mybuttons |= NOBUTTON;  /* right shift */

#if 0   /* needs 102 key keyboard (Bochs does not simulate one). */
      if ( (shift1 & 8) && (!(shift2 & 2)) )
	  mybuttons |= CALBUTTON;       /* alt, but not left one */
      if ( (shift1 & 4) && (shift2 & 1) )
	  mybuttons |= NEXTBUTTON;      /* ctrl, left one */
#else
      if ( shift1 & 8 )
	  mybuttons |= CALBUTTON;       /* any alt */
      if ( shift1 & 4 )
	  mybuttons |= NEXTBUTTON;      /* any ctrl */
#endif

  }


  if ( lpt_base > 0 ) {
      /* lpt_base > 0 means "use other than mouse for A/D"...   */
      int keybtn;       /* save */

      keybtn = mybuttons;
      out8 (lpt_base+2, in8 (lpt_base+2) | 0x23);
	  /* 1 "stb" (inverted on sub-d 25 - we want it OFF there)   */
	  /* 0x20 "direction: IN"   2 "afeed" (inverted on sub-d 25) */
      mybuttons = in8 (lpt_base) ^ lpt_inverse;
      mybuttons &= YESBUTTON | NOBUTTON | CALBUTTON | NEXTBUTTON; /* mask */

      if (mybuttons == (YESBUTTON | NOBUTTON | CALBUTTON | NEXTBUTTON) ) {
          lpt_inverse ^= YESBUTTON | NOBUTTON | CALBUTTON | NEXTBUTTON;
          mybuttons = 0;	/* assume all released, not all pressed */
      } /* other button polarity found -> auto-adjust dynamically */

      mybuttons |= keybtn;      /* restore / OR */
  }

  *bits = mybuttons;
  return get_ytime();

} /* get_buttons */



uint32 debounce(int buttons, int msecs) /* wait until all  of the */
{  /* given buttons have stayed unpressed for at least msecs...   */
   uint32 first_time, second_time, debounce_time;
   int my_buttons;

   msecs = ((msecs > 10000) || (msecs < 1)) ? 10000 : msecs; /* sanitize */
   debounce_time = (uint32)msecs * 1000;        /* 1000ysec/msec */
   do {
       do { /* wait until all desired buttons released for the first time */
	   first_time = get_buttons (&my_buttons);
       } while (buttons & my_buttons);

       do { /* loop until success or button pressed again */
	   second_time = get_buttons (&my_buttons);
       } while ( ((second_time-first_time) < debounce_time) ||
	   (buttons & my_buttons) );
   } while (buttons & my_buttons); /* repeat if still bouncing */

  return get_ytime ();

} /* debounce */

